@@ -4,15 +4,17 @@
 # This code is public domain.
 #
 
-from typing import Optional, Union, Callable, Dict, Any
+from typing import Optional, Union, Callable, Dict, List, Any
 import functools
-from flask import Flask, request, Response
-
+import inspect
 import datetime as dt
+
+from flask import Flask, request, Response
 from passlib.context import CryptContext  # type: ignore
 
 import logging
 log = logging.getLogger('auth')
+log.setLevel(logging.INFO)
 
 
 # carry data for error Response
@@ -33,6 +35,8 @@ CONF: Optional[Dict[str, Any]] = None
 # auth type
 AUTH: Optional[str] = None
 LAZY: Optional[bool] = None
+ALWAYS: Optional[bool] = False
+skip_path: Optional[List[Callable]] = None
 
 # auth token
 NAME: Optional[str] = None
@@ -59,10 +63,26 @@ UserInGroupType = Optional[Union[Callable[[str, str], bool],
                                  Callable[[str, int], bool]]]
 user_in_group: UserInGroupType = None
 
-# local copy of authenticated user
+# current authenticated user
 USER: Optional[str] = None
 
 
+# set, or possibly just reset, the current authentication
+def auth_set_user():
+    global USER
+    USER = None
+    if not ALWAYS:
+        return
+    for skip in skip_path:
+        if skip(request.path):
+            return
+    try:
+        USER = get_user()
+    except AuthException as e:
+        return e.message, e.status
+    assert USER is not None
+
+
 # wipe out current authentication
 def auth_after_cleanup(res: Response):
     global USER
@@ -70,32 +90,31 @@ def auth_after_cleanup(res: Response):
     return res
 
 
-def auth_before_cleanup():
-    global USER
-    USER = None
-
-
 # initialize module
 def setConfig(app: Flask,
               gup: GetUserPasswordType = None,
               uig: UserInGroupType = None):
-    global APP, CONF, AUTH, LAZY
+    global APP, CONF, AUTH, LAZY, ALWAYS, skip_path
     global NAME, REALM, SECRET, DELAY, GRACE, HASH, SIGLEN
     global LOGIN, USERP, PASSP, PM
     global get_user_password, user_in_group
     # overall setup
     APP = app
     CONF = app.config
-    app.before_request(auth_before_cleanup)
-    app.after_request(auth_after_cleanup)
+    # auth setup
     AUTH = CONF.get("FSA_TYPE", "httpd")
     LAZY = CONF.get("FSA_LAZY", True)
+    ALWAYS = CONF.get("FSA_ALWAYS", True)
+    app.before_request(auth_set_user)
+    app.after_request(auth_after_cleanup)
+    import re
+    skip_path = [re.compile(r).match for r in CONF.get("FSA_SKIP_PATH", [])]
     # token setup
     NAME = CONF.get("FSA_TOKEN_NAME", "auth")
-    import re
     realm = CONF.get("FSA_TOKEN_REALM", app.name).lower()
     # tr -cd "[a-z0-9_]" "": is there a better way to do that?
-    REALM = "".join(c for c in realm if re.match("[-a-z0-9_]", c))
+    keep_char = re.compile(r"[-a-z0-9_]").match
+    REALM = "".join(c for c in realm if keep_char(c))
     import random
     import string
     # list of 94 chars, about 6.5 bits per char
@@ -142,7 +161,6 @@ def get_fake_auth():
     # it could check that the user exists in db
     if user is None:
         raise AuthException("missing login parameter", 401)
-    log.info(f"LOGIN (param): {user}")
     return user
 
 
@@ -192,7 +210,6 @@ def get_basic_auth():
     if not request.is_secure:
         log.warning("password authentication over an insecure request")
     check_db_password(user, pwd)
-    log.info(f"LOGIN (basic): {user}")
     return user
 
 
@@ -215,7 +232,6 @@ def get_param_auth():
     if not request.is_secure:
         log.warning("password authentication over an insecure request")
     check_db_password(user, pwd)
-    log.info(f"LOGIN (param): {user}")
     return user
 
 
@@ -268,20 +284,19 @@ def get_token_auth(token):
     realm, user, limit, sig = token.split(':', 3)
     # check realm
     if realm != REALM:
-        log.info(f"LOGIN (token): unexpected realm {realm}")
+        log.debug(f"LOGIN (token): unexpected realm {realm}")
         raise AuthException(f"unexpected realm: {realm}", 401)
     # check signature
     ref = compute_signature(f"{realm}:{user}:{limit}", SECRET)
     if ref != sig:
-        log.info("LOGIN (token): invalid signature")
+        log.debug("LOGIN (token): invalid signature")
         raise AuthException("invalid auth token signature", 401)
     # check limit with a grace time
     now = get_timestamp(dt.datetime.utcnow() - dt.timedelta(minutes=GRACE))
     if now > limit:
-        log.info("LOGIN (token): token {token} has expired")
+        log.debug("LOGIN (token): token {token} has expired")
         raise AuthException("expired auth token", 401)
     # all is well
-    log.info(f"LOGIN (token): {user}")
     return user
 
 
@@ -300,12 +315,11 @@ def get_user():
 
     if AUTH == "httpd":
 
-        log.info(f"LOGIN (httpd): {request.remote_user}")
         USER = request.remote_user
 
     elif AUTH in ("fake", "param", "basic", "token", "password"):
 
-        # check for token
+        # always check for token
         if SECRET is not None and SECRET != "":
             params = request.values if request.json is None else request.json
             token = params.get(NAME, None)
@@ -328,13 +342,14 @@ def get_user():
             else:
                 raise AuthException("auth token is required", 401)
 
-        assert USER is not None  # else an exception would have been raised
-        return USER
-
     else:
 
         raise AuthException(f"unexpected authentication type: {AUTH}", 500)
 
+    assert USER is not None  # else an exception would have been raised
+    log.info(f"get_user({AUTH}): {USER}")
+    return USER
+
 
 #
 # authorize decorator
@@ -370,57 +385,101 @@ class authorize:
         return wrapper
 
 
-#
-# parameters decorator
-#
-class parameters:
+def bool_cast(s: str) -> Optional[bool]:
+    return None if s is None else \
+        False if s.lower() in ("", "0", "false") else \
+        True
 
-    def __init__(self, *args, **kwargs):
-        self.params = args
-        self.typed = kwargs
 
-    def __call__(self, fun):
+def int_cast(s: str) -> Optional[int]:
+    return None if s is None else int(s, base=0)
+
+
+# note: mypy complains wrongly about non-existing _empty.
+CASTS = {bool: bool_cast, int: int_cast, inspect._empty: str}
+
+
+#
+# autoparams decorator
+#
+# required:
+# - None: function parameters are required unless there is a default value
+# - True: all function parameters are required
+# - False: all function parameters are optional,
+#   with default value None unless explicitely provided
+#
+# allparams:
+# - whether all request parameters are automatically translated to function
+#   parameters with a str value.
+#
+# others:
+# - args: list of expected parameters, implicit type is str
+# - kwargs: list of expected parameters, explicit type as a value
+#
+def parameters(*args, required=None, allparams=False, **kwargs):
+
+    def decorate(fun):
+
+        types: Dict[str, Callable] = {}
+        defaults: Dict[str, Any] = {}
+
+        # parameters types from **kwargs
+        for n, t in kwargs.items():
+            if n not in types:
+                types[n] = CASTS.get(t, t)
+            else:
+                log.warning(f"ignoring *kwargs decorator parameter {n}")
+
+        # parameters from *args
+        for n in args:
+            if n not in types:
+                types[n] = str
+            else:
+                log.warning(f"ignoring *args decorator parameter {n}")
+
+        # parameters types/casts and defaults from signature
+        sig = inspect.signature(fun)
+
+        for n, p in sig.parameters.items():
+            if n not in types and \
+               p.kind not in (p.VAR_KEYWORD, p.VAR_POSITIONAL):
+                types[n] = CASTS.get(p.annotation, p.annotation)
+            if p.default != inspect._empty:
+                defaults[n] = p.default
+
         @functools.wraps(fun)
         def wrapper(*args, **kwargs):
+
+            # translate request parameters to named function parameters
             params = request.values if request.json is None else request.json
-            for p in self.params:
-                if p not in params:
-                    return f"missing mandatory parameter: {p}", 400
-                kwargs[p] = params[p]
-            for p in self.typed:
-                if p not in params:
-                    return f"missing mandatory parameter: {p}", 400
-                try:
-                    kwargs[p] = self.typed[p](params[p])
-                except Exception as e:
-                    log.debug(f"type error on {p}: {e}")
-                    return f"type error on parameter {p} ({e})", 400  # 422?
-            # ok to proceed
-            return fun(*args, **kwargs)
-        return wrapper
-
-
-class autoparams:
-
-    def __init__(self, required=True):
-        self.required = required
-
-    def __call__(self, fun):
-        @functools.wraps(fun)
-        def wrapper(*args, **kwargs):
-            params = request.values if request.json is None else request.json
-            for p, typing in fun.__annotations__.items():
-                if p == "return":
-                    continue
-                if p not in params and p not in kwargs:
-                    if self.required:
-                        return f"missing parameter {p}", 400
+            for p, typing in types.items():
+                # guess which function parameters are request parameters
+                if p not in kwargs:
+                    if p in params:
+                        try:
+                            kwargs[p] = typing(params[p])
+                        except Exception as e:
+                            return f"type error on parameter {p} ({e})", 400
                     else:
-                        kwargs[p] = None
-                if p in params and p not in kwargs:
-                    try:
-                        kwargs[p] = typing(params[p])
-                    except Exception as e:
-                        return f"type error on parameter {p} ({e})", 400
+                        if required is None:
+                            if p in defaults:
+                                kwargs[p] = defaults[p]
+                            else:
+                                return f"missing parameter {p}", 400
+                        elif required:
+                            return f"missing parameter {p}", 400
+                        else:
+                            kwargs[p] = defaults.get(p, None)
+
+            # possibly add others, without shadowing already provided ones
+            if allparams:
+                for p in params:
+                    if p not in kwargs:
+                        kwargs[p] = params[p]
+
+            # then call the initial function
             return fun(*args, **kwargs)
+
         return wrapper
+
+    return decorate
